#INCLUDE "simpleio.CH"
#INCLUDE "dcdialog.CH"
#INCLUDE "asxml.CH"
#INCLUDE "dcads.CH"

#Pragma Library("asxml10.lib")
#Pragma Library("dclipx.lib")


#DEFINE CRLF Chr(13)+Chr(10)

STATIC aText[0]

FUNCTION DC_Xml2ObjectTree( cXmlFile, lXb2Net )

LOCAL nXMLDoc, nActions, oRootNode, nTag, nHandle, nOutHandle, cXmlText, ;
      cString, lIsString, aXml

DEFAULT lXb2Net := .f.

IF Empty(cXmlFile) .OR. Valtype(cXmlFile) # 'C'
  RETURN nil
ENDIF

IF '<' $ cXmlFile
  cXmlText := cXmlFile
  lIsString := .T.
ELSE
  cXmlText := MemoRead(cXmlFile)
  lIsString := .F.
ENDIF

IF "'" $ cXmlText .AND. !lIsString
  nHandle := _TxtOpen(cXmlFile)
  nOutHandle := FCreate(cXmlFile + '.TEMP')
  DO WHILE !_TxtEof(nHandle)
    cString := _TxtLine(nHandle)
    _TxtSkip(nHandle,1)
    cString := Strtran(cString,"'",Chr(255))
    FWrite(nOutHandle,cString+Chr(13)+Chr(10))
  ENDDO
  FClose(nOutHandle)
  FClose(nHandle)
  cXmlFile += '.TEMP'
ENDIF

IF lXb2Net

  aXml := &('xbXML2Array')(cXmlText)

  oRootNode := DC_XmlNode():New('#ROOT')
  DumpTag2( oRootNode, aXml )

ELSE

  IF lIsString
    nXMLDoc := XMLDocOpenString(cXmlText)
  ELSE
    nXMLDoc := XMLDocOpenFile(cXmlFile)
  ENDIF

  IF(nXMLDoc == 0)
    PrintErrors(nXMLDoc)
    RETURN nil
  ENDIF

  nTag := XMLDocGetRootTag(nXMLDoc)

  oRootNode := DumpTag(nTag,1)

  XMLDocClose(nXMLDoc)

ENDIF

IF '.TEMP' $ cXmlFile
  Ferase(cXmlFile)
ENDIF

RETURN oRootNode

* -------------

STATIC FUNCTION DumpTag(nTag,nLevel)

LOCAL aMember, n, oNode, oChildNode

IF !XMLGetTag(nTag, @aMember)
  RETURN nil
ENDIF

IF aMember[XMLTAG_NAME] == '#CDATA'
  IF !Empty(aMember[XMLTAG_CHILD])
    nTag := aMember[XMLTAG_CHILD][1]
    IF !XMLGetTag(nTag,@aMember)
      RETURN nil
    ENDIF
  ELSE
    // RETURN nil
  ENDIF
ELSEIF aMember[XMLTAG_NAME] == '#PCDATA'
  IF !Empty(aMember[XMLTAG_CHILD])
    nTag := aMember[XMLTAG_CHILD][1]
    IF !XMLGetTag(nTag,@aMember)
      RETURN nil
    ENDIF
  ELSE
    RETURN nil
  ENDIF
ENDIF

oNode := DC_XmlNode():new(aMember[XMLTAG_NAME],aMember[XMLTAG_CONTENT], ;
                          aMember[XMLTAG_ATTRIB],aMember[XMLTAG_CHILD])

Sleep(0)

IF aMember[XMLTAG_CHILD] != NIL
  FOR n := 1 TO Len(aMember[XMLTAG_CHILD])
    oChildNode := DumpTag(aMember[XMLTAG_CHILD][n],nLevel++)
    IF !oChildNode == NIL
      oNode:addChild(oChildNode)
    ENDIF
  NEXT
ENDIF

RETURN oNode

* ------------

STATIC FUNCTION DumpTag2( oNode, aXml )

LOCAL oChildNode, i

IF Valtype(aXml) == 'A'
  FOR i := 1 TO Len(aXml)
    IF Valtype(aXml[i,3]) == 'A'
      oChildNode := DC_XmlNode():new(aXml[i,1],nil,aXml[i,2])
      oNode:addChild(oChildNode)
      DumpTag2(oChildNode,aXml[i,3])
    ELSE
      oChildNode := DC_XmlNode():new(aXml[i,1],aXml[i,3],aXml[i,2])
      oNode:addChild(oChildNode)
    ENDIF
  NEXT
ELSE
  oChildNode := DC_XmlNode():new(aXml[1],aXml[3],aXml[2])
  oNode:addChild(oChildNode)
ENDIF

RETURN oNode

* -------------

STATIC FUNCTION DumpTagArray(nTag,nLevel)

LOCAL aMember, n, aArray, aChildArray

IF !XMLGetTag(nTag, @aMember)
  RETURN nil
ENDIF

IF aMember[XMLTAG_NAME] == '#CDATA'
  IF !Empty(aMember[XMLTAG_CHILD])
    nTag := aMember[XMLTAG_CHILD][1]
    IF !XMLGetTag(nTag,@aMember)
      RETURN nil
    ENDIF
  ELSE
    RETURN nil
  ENDIF
ELSEIF aMember[XMLTAG_NAME] == '#PCDATA'
  IF !Empty(aMember[XMLTAG_CHILD])
    nTag := aMember[XMLTAG_CHILD][1]
    IF !XMLGetTag(nTag,@aMember)
      RETURN nil
    ENDIF
  ELSE
    RETURN nil
  ENDIF
ENDIF

aArray := { aMember[XMLTAG_NAME],aMember[XMLTAG_CONTENT], ;
            aMember[XMLTAG_ATTRIB] }

Sleep(0)

IF aMember[XMLTAG_CHILD] != NIL
  FOR n := 1 TO Len(aMember[XMLTAG_CHILD])
    aChildArray := DumpTagArray(aMember[XMLTAG_CHILD][n])
    IF !aChildArray == NIL
      AAdd( aArray, aChildArray )
    ENDIF
  NEXT
ENDIF

RETURN aArray

* -----------

STATIC PROCEDURE PrintErrors(nXmlDoc)

LOCAL n, aErrors := XMLDocGetErrorList(nXMLDoc)

    IF aErrors == NIL
        WTF "No error descriptions available!"
    ELSEIF Valtype(aErrors) == 'A'
      FOR n := 1 TO Len(aErrors)
         WTF aErrors[n, XML_ERROR_FILE]+;
          "("+AllTrim(Str(aErrors[n, XML_ERROR_LINE]))+":"+;
          AllTrim(Str(aErrors[n, XML_ERROR_COLUMN]))+")"+;
          " error XPL00"+AllTrim(Str(aErrors[n, XML_ERROR_ID]))+": "+;
          GetErrorText(aErrors[n, XML_ERROR_ID])
        IF aErrors[n, XML_ERROR_ADDINFO] != NIL
            WTF aErrors[n, XML_ERROR_FILE]+;
              "("+AllTrim(Str(aErrors[n, XML_ERROR_LINE]))+":"+;
              AllTrim(Str(aErrors[n, XML_ERROR_COLUMN]))+")"+;
              " info: "+aErrors[n, XML_ERROR_ADDINFO]
        ENDIF
      NEXT
    ENDIF

RETURN

* ---------------

STATIC FUNCTION GetErrorText(nError)

STATIC aErrorText := {;
        "out of memory; close some applications and try again",;
        "invalid DTD declaration",;
        "content outside XML tags not allowed",;
        "document has no root tag",;
        "invalid endtag or no endtag found",;
        "expecting string delimiter",;
        "unterminated string",;
        "file not found",;
        "file read error",;
        "missing name of file",;
        "duplicate attribute",;
        "malformed attribute",;
        "invalid name for attribute",;
        "error while processing action codeblock",;
    }

RETURN aErrorText[nError]

* -----------------

CLASS DC_XmlNode

   EXPORTED:

   VAR name
   VAR content
   VAR attrib
   VAR childList
   VAR childTags
   VAR data
   VAR parent
   VAR cargo
   VAR caption

   METHOD Init, AddChild, AddNode, ChildList, Status, FindNode, GetAttr, ;
          TextDump, WriteXML, WriteNode, AsString, NodeToArray

ENDCLASS

* ---------------

METHOD DC_XmlNode:Init( cName, xContent, aAttrib, aChildTags, aData )

LOCAL i, aAttrib2

IF Valtype(aAttrib) == 'C'
  aAttrib := Strtran(aAttrib,Chr(13)+Chr(10),'')
   aAttrib := _TokenArray(aAttrib,' ',.t.)
  FOR i := 1 TO Len(aAttrib)
    IF Empty(aAttrib[i])
      ARemove(aAttrib,i)
      i--
      LOOP
    ENDIF
    aAttrib2 := _TokenArray(aAttrib[i],'=',.t.)
    aAttrib[i] := aAttrib2
  NEXT
ENDIF

::name := cName
::content := xContent
IF Valtype(::content) = 'C' .AND. Chr(255) $ ::content
  ::content := Strtran(::content,Chr(255),"'")
ENDIF
::attrib := aAttrib
::childList := {}
::childTags := aChildTags
::data := aData
::caption := ::name

RETURN self

* ---------------

METHOD DC_XmlNode:AddChild( oNode )
AAdd(::childList, oNode)
oNode:parent := self
RETURN self

* ---------------

METHOD DC_XmlNode:AddNode( oNode )
AAdd(::childList, oNode)
oNode:parent := self
RETURN self

* ---------------

METHOD DC_XmlNode:Childlist()
RETURN ::childList

* ---------------

METHOD DC_XmlNode:Status()
RETURN 1

* ---------------

METHOD DC_XmlNode:FindNode( acTagName, lMultiple, lSearchSubs )

LOCAL oNode, i, aNodes[0], cTagName, nStart := 1

DEFAULT lMultiple := .F., ;
        lSearchSubs := .f.

oNode := self

IF Valtype(acTagName) == 'A'

  IF DC_XmlNodeFindMode() == 1 .AND. acTagName[1] == self:name
    nStart := 2
  ENDIF

  FOR i := nStart TO Len(acTagName)
    cTagName := acTagName[i]
    oNode := oNode:findNode( cTagName, IIF(i==Len(acTagName),lMultiple,.f.), .t. )
    IF Valtype(oNode) == 'O'
      IF oNode:name == ATail(acTagName)
        RETURN oNode
      ENDIF
    ELSEIF Valtype(oNode) == 'A'
      RETURN oNode
    ELSE
      RETURN nil
    ENDIF
  NEXT
  RETURN nil
ENDIF

FOR i := 1 TO Len(::childList)
  oNode := ::childList[i]
  IF oNode:caption == acTagName .OR. Substr(oNode:caption,At(':',oNode:caption)+1) == acTagName
    IF lMultiple
      AAdd(aNodes,oNode)
    ELSE
      RETURN oNode
    ENDIF
  ELSEIF lSearchSubs
    oNode := oNode:findNode( acTagName,,.t.)
    IF Valtype(oNode) == 'O'
      IF lMultiple
        AAdd(aNodes,oNode)
      ELSE
        RETURN oNode
      ENDIF
    ENDIF
  ENDIF
NEXT

IF Empty(aNodes)
  aNodes := NIL
ENDIF

RETURN aNodes

* ----------

METHOD DC_XmlNode:GetAttr( cAttrName, lRemoveQuotes )

LOCAL aAttr, i, cAttr
DEFAULT lRemoveQuotes := .f.

IF Empty(::attrib)
  RETURN nil
ENDIF
FOR i := 1 TO Len(::attrib)
  IF ::attrib[i,1] == cAttrName
    cAttr := Strtran(::attrib[i,2],Chr(255),"'")
    IF lRemoveQuotes
      IF (Left(cAttr,1) == "'" .AND. Right(cAttr,1) == "'") .OR. ;
         (Left(cAttr,1) == '"' .AND. Right(cAttr,1) == '"')
        cAttr := SubStr(cAttr,2)
        cAttr := SubStr(cAttr,1,Len(cAttr)-1)
      ENDIF
    ENDIF
    RETURN cAttr
  ENDIF
NEXT
RETURN nil

* ---------------

METHOD DC_XmlNode:TextDump( nIndent, cDelim, nLevel )

LOCAL cOutString := '', i

DEFAULT nIndent := 3, ;
        cDelim := ' : ', ;
        nLevel := 0

cOutString += ( Space(nIndent*nLevel) + ::caption + cDelim + _XtoC(::content) + CRLF )
IF Valtype(::attrib) == 'A'
  FOR i := 1 TO Len(::attrib)
    cOutString += ( Space(nIndent*(nLevel+1)) + ::attrib[i,1] + cDelim + _XtoC(::attrib[i,2]) + CRLF )
  NEXT
ENDIF
FOR i := 1 TO Len(::childList)
  cOutString += ::childList[i]:TextDump( nIndent, cDelim, nLevel+1)
NEXT
cOutString := Strtran(cOutString,Chr(255),"'")
RETURN cOutString

* ----------------

METHOD DC_XmlNode:WriteXML( cFileName, lCanon, lSuppressIndent )

LOCAL nHandle
nHandle := FCreate(cFileName)
IF nHandle <= 0
  RETURN .F.
ENDIF
FWrite(nHandle,'<?xml version="1.0" encoding="UTF-8" ?>'+CRLF)
::writeNode( nHandle,lCanon,lSuppressIndent)
FClose(nHandle)

RETURN .T.

* ----------------

METHOD DC_XmlNode:WriteNode( nHandle, lCanon, lSuppressIndent, nIndent )

LOCAL i, cAttrib := '', lNoIndent := .F., lCloseTag, cLineEnd

DEFAULT nIndent := 0, ;
        lCanon := .f.,;
        lSuppressIndent := .F.

IF lCanon
  cLineEnd := chr(10)
ELSE
  cLineEnd := CRLF
ENDIF

IF !Empty(::attrib)
  IF lCanon
    ASort(::attrib,,,{|a,b|a[1]<b[1]})
  ENDIF
  cAttrib := ' '
  FOR i := 1 TO Len(::attrib)
    cAttrib += ::attrib[i,1] + '="' + Trim(_XtoC(::attrib[i,2])) + '" '
  NEXT
  cAttrib := left(cAttrib,len(cAttrib)-1)
ENDIF

lCloseTag := !(Empty(::childList) .AND. ;
               Empty(::data) .AND. ;
               Empty(::content))

FWrite( nHandle, Space(nIndent) + '<' + ::name + cAttrib + ;
                 iif(Empty(::childList) .AND. Empty(::data) .AND. Empty(::content), ;
                     iif(lCanon,'></' + ::name,'/'),'') +  ;
                 '>' + IIF(Empty(::content), cLineEnd, ''))

IF !Empty(::content)
  FWrite( nHandle, Trim(_XtoC(::content)) )
  lNoIndent := .T.
ENDIF

IF lSuppressIndent
  nIndent += 0
ELSEIF !lCanon
  nIndent += 2
ELSE
  nIndent += 2
ENDIF

IF !Empty(::data)
  lNoIndent := .F.
  FOR i := 1 TO Len(::data)
    FWrite( nHandle, Space(nIndent) + '<' + ::data[i,1] + '>' + ;
                     Trim(_XtoC(::data[i,2])) + '</' + ::data[i,1] + '>' + cLineEnd )
  NEXT
ENDIF

FOR i := 1 TO Len(::childList)
  lNoIndent := .F.
  ::childList[i]:writeNode( nHandle, lCanon, lSuppressIndent, nIndent )
NEXT

IF lSuppressIndent
  nIndent := 0
ELSEIF lNoIndent
  nIndent := 0
ELSE
  nIndent -= 2
ENDIF

IF lCloseTag
  FWrite( nHandle, Space(nIndent) + '</' + ::name + '>' + cLineEnd )
ENDIF

RETURN .T.

* ---------------

METHOD DC_XmlNode:AsString( lCanon, lSuppressIndent, nIndent )

LOCAL i, cAttrib := '', lNoIndent := .F., lCloseTag, cString := '', cVarName, xValue, cLineEnd

DEFAULT nIndent := 0, ;
        lCanon := .F.,;
        lSuppressIndent := .F.

IF lCanon
  cLineEnd := chr(10)
ELSE
  cLineEnd := CRLF
ENDIF

IF !Empty(::attrib)
  IF lCanon
    ASort(::attrib,,,{|a,b|a[1]<b[1]})
  ENDIF
  cAttrib := ' '
  FOR i := 1 TO Len(::attrib)
    cAttrib += ::attrib[i,1] + '="' + Trim(_XtoC(::attrib[i,2])) + '" '
  NEXT
  cAttrib := left(cAttrib,len(cAttrib)-1)
ENDIF

lCloseTag := !(Empty(::childList) .AND. ;
               Empty(::data) .AND. ;
               Empty(::content))

cString += Space(nIndent) + '<' + ::name + cAttrib + ;
           iif(Empty(::childList) .AND. Empty(::data) .AND. Empty(::content), ;
               iif(lCanon,'></' + ::name,'/'),'') +  ;
           '>' + IIF(Empty(::content), cLineEnd, '')

IF !Empty(::content)
  cString += Trim(_XtoC(::content))
  lNoIndent := .T.
ENDIF

IF lSuppressIndent
  nIndent += 0
ELSEIF !lCanon
  nIndent += 2
ELSE
  nIndent += 2
ENDIF

IF !Empty(::data)
  lNoIndent := .F.
  FOR i := 1 TO Len(::data)
    cString += Space(nIndent) + '<' + ::data[i,1] + '>' + ;
               Trim(_XtoC(::data[i,2])) + '</' + ::data[i,1] + '>' + cLineEnd
  NEXT
ENDIF

FOR i := 1 TO Len(::childList)
  lNoIndent := .F.
  cString += ::childList[i]:asString( lCanon, lSuppressIndent, nIndent )
NEXT

IF  lSuppressIndent
  nIndent := 0
ELSEIF lNoIndent
  nIndent := 0
ELSE
  nIndent -= 2
ENDIF

IF lCloseTag
  cString += Space(nIndent) + '</' + ::name + '>' + cLineEnd
ENDIF

RETURN cString

* ---------

METHOD DC_XmlNode:NodeToArray( aArray, lAttribSubArray )

LOCAL i, aChildList := ::childList(), aNodes

DEFAULT lAttribSubArray := .F.

IF !Empty(aChildList)
  aNodes := Array(Len(aChildList))
ENDIF
FOR i := 1 TO Len(aChildList)
  aNodes[i] := Array(0)
  aChildList[i]:nodeToArray(aNodes[i])
NEXT

ASize(aArray,2)
DEFAULT ::name := '', ;
        ::content := ''

aArray[1] := ::name
aArray[2] := IIF(!Empty(aNodes),aNodes,::content)
IF !Empty(::attrib)
  FOR i := 1 TO Len(::attrib)
    IF lAttribSubArray
      AAdd(aArray,::attrib[i])
    ELSE
      ASize(aArray,Len(aArray)+2)
      aArray[Len(aArray)-1] := ::attrib[i,1]
      aArray[Len(aArray)] := ::attrib[i,2]
    ENDIF
  NEXT
ENDIF

RETURN self

* -------------

FUNCTION DC_WorkArea2Xml( cXmlFile, aStruct, aTags )

LOCAL oRoot, oStructure, oData, oField, oFieldName, oFieldType, i, ;
      oFieldLen, oFieldDec, nSaveRec := RecNo(), oRecord, cFieldName, ;
      xValue, aTagDefault

aTagDefault := { ;
   'Root', ;
   'Structure', ;
   'Field', ;
   'FieldName', ;
   'FieldType', ;
   'FieldLen', ;
   'FieldDec', ;
   'Data', ;
   'Record' }

DEFAULT aStruct := dbStruct(), ;
        aTags := aTagDefault

oRoot := DC_XmlNode():new(aTags[1])

oStructure := DC_XmlNode():new(aTags[2])

oRoot:addNode(oStructure)

FOR i := 1 TO Len(aStruct)
  oField := DC_XmlNode():new(aTags[3])
  oStructure:addNode(oField)
  oFieldName := DC_XmlNode():new(aTags[4],aStruct[i,1])
  oField:addNode(oFieldName)
  oFieldType := DC_XmlNode():new(aTags[5],_TypeDesc(aStruct[i,2]))
  oField:addNode(oFieldType)
  oFieldLen := DC_XmlNode():new(aTags[6],aStruct[i,3])
  oField:addNode(oFieldLen)
  oFieldDec := DC_XmlNode():new(aTags[7],aStruct[i,4])
  oField:addNode(oFieldDec)
NEXT

oData := DC_XmlNode():new(aTags[8])
oRoot:addNode(oData)

DC_DbGoTop()

DO WHILE !DC_Eof()
  oRecord := DC_XmlNode():new(aTags[9])
  oData:addNode(oRecord)
  FOR i := 1 TO Len(aStruct)
    cFieldName := aStruct[i,1]
    xValue := &(cFieldName)
    oField := DC_XmlNode():new(cFieldName,xValue)
    oRecord:addNode(oField)
  NEXT
  DC_DbSkip()
ENDDO

DC_DbGoTo(nSaveRec)

IF !Empty(cXmlFile)
  RETURN oRoot:writeXml(cXmlFile)
ENDIF

RETURN oRoot:asString()

* --------------

STATIC FUNCTION _TypeDesc( cType )

LOCAL cDesc

IF cType = 'C'
  cDesc := 'Character'
ELSEIF cType = 'N'
  cDesc := 'Numeric'
ELSEIF cType = 'D'
  cDesc := 'Date'
ELSEIF cType = 'L'
  cDesc := 'Logical'
ELSEIF cType = 'F'
  cDesc := 'Double'
ELSEIF cType = 'X'
  cDesc := 'Character/Binary'
ELSEIF cType = 'O'
  cDesc := 'Generic'
ELSEIF cType = 'I'
  cDesc := 'Integer'
ELSEIF cType = 'M'
  cDesc := 'Memo/Text'
ELSEIF cType = 'V'
  cDesc := 'Memo/Binary'
ELSEIF cType = 'T'
  cDesc := 'TimeStamp'
ELSEIF cType = 'Y'
  cDesc := 'Currency'
ELSE
  cDesc := 'Unknown'
ENDIF

RETURN cDesc

* --------------

FUNCTION DC_Xml2WorkArea( cXml )

LOCAL oRoot, aRecords, oRecord, oRecordNode

oRoot := DC_Xml2ObjectTree( cXml )

aRecords := oRoot:findNode({'Root','Data','Record'},.t.,.t.)

oRecord := DC_DbRecord():new()

wtf aRecords, Len(aRecords)

FOR i := 1 TO Len(aRecords)
  oRecordNode := aRecords[i]
  aFieldNodes := oRecordNode:childList()
  FOR j := 1 TO Len(aFieldNodes)
    cFieldName := aFieldNodes[j]:name
    cValue := aFieldNodes:content
  NEXT
NEXT

RETURN nil

* --------------

STATIC FUNCTION _AdsTypeDesc( nType )

LOCAL cDesc := 'Unknown'

IF nType == ADS_LOGICAL
  cDesc := 'Logical'
ELSEIF nType == ADS_NUMERIC
  cDesc := 'Numeric'
ELSEIF nType == ADS_DATE
  cDesc := 'Date'
ELSEIF nType == ADS_STRING
  cDesc := 'Character'
ELSEIF nType == ADS_MEMO
  cDesc := 'Memo/Text'
ELSEIF nType == ADS_BINARY
  cDesc := 'Memo/Binary'
ELSEIF nType == ADS_IMAGE
  cDesc := 'Bitmap'
ELSEIF nType == ADS_VARCHAR
  cDesc := 'VarChar'
ELSEIF nType == ADS_COMPACTDATE
  cDesc := 'CompactDate'
ELSEIF nType == ADS_DOUBLE
  cDesc := 'Double'
ELSEIF nType == ADS_INTEGER
  cDesc := 'Integer'
ELSEIF nType == ADS_SHORTINT
  cDesc := 'ShortInteger'
ELSEIF nType == ADS_TIME
  cDesc := 'Time'
ELSEIF nType == ADS_TIMESTAMP
  cDesc := 'TimeStamp'
ELSEIF nType == ADS_AUTOINC
  cDesc := 'AutoIncrement'
ELSEIF nType == ADS_RAW
  cDesc := 'Raw'
ELSEIF nType == ADS_CURDOUBLE
  cDesc := 'CurrencyDouble'
ELSEIF nType == ADS_MONEY
  cDesc := 'Money'
ELSEIF nType == ADS_LONGLONG
  cDesc := 'LongLong'
ELSEIF nType == ADS_CISTRING
  cDesc := 'CIString'
ENDIF

RETURN cDesc

* -----------------

FUNCTION DC_AdsCursor2XML( nCursorHandle, cXMLFile, cSelect )

LOCAL nNumFields := 0, i, cFieldName, nLength, lOk, nType, nLen, nDec, ;
      aData[0], aStruct[0], aColumn, nDidSkip, xValue, GetList[0], ;
      aTags, oRoot, oStructure, oFieldName, oFieldType, ;
      oFieldLen, oFieldDec, oRecord, oField, oData

aTags := { ;
   'Root', ;
   'Structure', ;
   'Field', ;
   'FieldName', ;
   'FieldType', ;
   'FieldLen', ;
   'FieldDec', ;
   'Data', ;
   'Record' }

AdsGoToTop(nCursorHandle)

nNumFields := 0
AdsGetNumFields( nCursorHandle, @nNumFields )

aColumn := Array(nNumFields)
FOR i := 1 TO nNumFields
  cFieldName := Space(20)
  nLength := 100
  AdsGetFieldName( nCursorHandle, i, @cFieldName, @nLength )
  cFieldName := Alltrim(Strtran(cFieldName,Chr(0),''))
  nType := 0
  AdsGetFieldType( nCursorHandle, cFieldName, @nType )
  nLen := 0
  AdsGetFieldLength( nCursorHandle, cFieldName, @nLen )
  nDec := 0
  AdsGetFieldDecimals( nCursorHandle, cFieldName, @nDec )
  AAdd( aStruct, { cFieldName, ;
                   _AdsTypeDesc(nType), ;
                   nLen, ;
                   nDec, ;
                   DC_AdsFieldBlock( nCursorHandle, cFieldName ) } )
NEXT

oRoot := DC_XmlNode():new(aTags[1])

oStructure := DC_XmlNode():new(aTags[2])

oRoot:addNode(oStructure)

oData := DC_XmlNode():new(aTags[8])
oRoot:addNode(oData)

FOR i := 1 TO Len(aStruct)
  oField := DC_XmlNode():new(aTags[3])
  oStructure:addNode(oField)
  oFieldName := DC_XmlNode():new(aTags[4],aStruct[i,1])
  oField:addNode(oFieldName)
  oFieldType := DC_XmlNode():new(aTags[5],_TypeDesc(aStruct[i,2]))
  oField:addNode(oFieldType)
  oFieldLen := DC_XmlNode():new(aTags[6],aStruct[i,3])
  oField:addNode(oFieldLen)
  oFieldDec := DC_XmlNode():new(aTags[7],aStruct[i,4])
  oField:addNode(oFieldDec)
NEXT

nDidSkip := 1
DO WHILE nDidSkip > 0
  nDidSkip := DC_AdsSkipper(nCursorHandle,1)
  oRecord := DC_XmlNode():new(aTags[9])
  oData:addNode(oRecord)
  FOR i := 1 TO Len(aStruct)
    cFieldName := aStruct[i,1]
    xValue := Eval(aStruct[i,5])
    oField := DC_XmlNode():new(cFieldName,xValue)
    oRecord:addNode(oField)
  NEXT
ENDDO

AdsGoToTop(nCursorHandle)

IF !Empty(cXmlFile)
  RETURN oRoot:writeXml(cXmlFile)
ENDIF

RETURN oRoot:asString()

* ------------

FUNCTION DC_XmlEscape( cText, lEscape )

DEFAULT lEscape := .t.

IF lEscape
  cText := Strtran( cText, '&', '&amp;' )
  cText := Strtran( cText, '<', '&lt;' )
  cText := Strtran( cText, '>', '&gt;' )
  cText := Strtran( cText, '"', '&quot;' )
  cText := Strtran( cText, "'", '&apos;' )
ELSE
  cText := Strtran( cText, '&amp;', '&' )
  cText := Strtran( cText, '&lt;', '<' )
  cText := Strtran( cText, '&gt;', '>' )
  cText := Strtran( cText, '&quot;', '"' )
  cText := Strtran( cText, '&apos;', "'" )
ENDIF

RETURN cText

* ------------

GETSETFUNCTION DC_XmlNodeFindMode DEFAULT 0

* ----------

STATIC FUNCTION _txtopen ( cFileName, nMode )

LOCAL nHandle := FOpen(cFileName, nMode)

IF nHandle > 0
  IF Len(aText) < nHandle
    ASize(aText,nHandle)
  ENDIF
  aText[nHandle] := 1
ENDIF

RETURN nHandle

* -----------------

STATIC FUNCTION _txteof ( nHandle )

LOCAL nEnd,nCurrent

DEFAULT nHandle := 0
IF nHandle <= 0
  RETURN .f.
ENDIF
nCurrent := FSeek(nHandle,0,1)
nEnd := FSeek(nHandle,0,2) - 1
FSeek(nHandle,nCurrent,0)
RETURN nCurrent>=nEnd

* -----------------

STATIC FUNCTION _txtskip ( nHandle, nLines, cEol, nLineLen )

LOCAL cString, cChar, nPos, nEnd, i

cEol := IIF( Valtype(cEol)='C',cEol,Chr(13)+Chr(10))

DEFAULT nHandle := 0, ;
        nLines := 1, ;
        nLineLen := 40000

IF nHandle <= 0
  RETURN 0
ENDIF
IF Len(aText)<nHandle
  ASize(aText,nHandle)
  aText[nHandle] := 0
ENDIF
FOR i := 1 TO nLines
  nPos := FSeek( nHandle, 0, 1 ) // Save current pointer
  cString := Space(nLineLen)
  FRead( nHandle, @cString, nLineLen )
  nEnd := At( cEol, cString ) // End of line
  IF nEnd=0
    nPos := FSeek( nHandle, 0, 2)
    EXIT
  ENDIF
  FSeek( nHandle, nPos, 0 ) // Go back to saved position
  nPos := FSeek( nHandle, nPos+nEnd, 0 ) // Go to start of next line
NEXT
aText[nHandle] += (i-1)
RETURN aText[nHandle]

* -----------------

STATIC FUNCTION _txtline ( nHandle, cEol, nLineLen )

LOCAL cString, cChar, nPos, nEnd, i

DEFAULT nLineLen := 40000

cEol := IIF( Valtype(cEol)='C',cEol,Chr(13)+Chr(10))

DEFAULT nHandle := 0
IF nHandle <= 0 .OR. Len(aText) < nHandle .OR. Valtype(aText[nHandle])#'N'
  RETURN ''
ENDIF

nPos := FSeek( nHandle, 0, 1 )
cString := Space(nLineLen)
FRead( nHandle, @cString, nLineLen )

nEnd := At( cEol, cString )
FSeek( nHandle, nPos, 0 )
IF nEnd=0
  RETURN cString
ENDIF
cString := SubStr( cString, 1, nEnd-1)
FOR i := 1 TO Len(cEol)
  cString := StrTran( cString,SubStr(cEol,i,1),'')
NEXT
RETURN cString

* --------

STATIC FUNCTION _XtoC( xVar, lString, nLength )

LOCAL cType := ValType(xVar), xRet

DEFAULT lString := .f.

DO CASE
      Case cType = "C" .or. cType="M"
         xRet := xVar
      Case cType = "N"
         xRet := Alltrim(Str(xVar))
      Case cType = "D"
         xRet := DtoC(xVar)
      Case cType = "L"
         xRet := If(xVar,"Yes","No ")
      Case cType = "U"
         xRet := ""
ENDCASE
Return xRet

* ------------

STATIC FUNCTION _tokenarray( cString, cDelims, lQuote )

LOCAL nFound, i, lStart, lDoubleQuoteOn, cInString, cChar, nCount := 0, ;
      aTokens[0], lSingleQuoteOn

DEFAULT lQuote := .f.

cDelims := IIF(Valtype(cDelims)='C',cDelims,_TokenDelims() )
IF EMPTY(cString)
  RETURN aTokens
ENDIF
IF lQuote
  lDoubleQuoteOn := .F.
  lSingleQuoteOn := .F.
  cInString := cString
  cString := ''
  FOR i := 1 TO Len(cInString)
    cChar := cInString[i]
    IF cChar == '"'
      lDoubleQuoteOn := !lDoubleQuoteOn
    ELSEIF cChar == "'"
      lSingleQuoteOn := !lSingleQuoteOn
    ELSEIF !lDoubleQuoteOn .AND. !lSingleQuoteOn .AND. cChar $ cDelims
      cChar := Chr(0)
    ENDIF
    cString += cChar
  NEXT
  cString := StrTran(cString,'"','')
ELSE
  FOR i := 1 TO LEN(cDelims)
    cString := Strtran(cString,SubStr(cDelims,i,1),CHR(0))
  NEXT
ENDIF
IF Len(cDelims) > 1
  IF CHR(0)+CHR(0) $ cString
    cString := StrTran(cString,CHR(0)+CHR(0),CHR(0))
  ENDIF
ENDIF
DO WHILE .t.
  nFound := AT(CHR(0),cString)
  IF nFound = 1 //.AND. Len(aTokens)=0
    AADD( aTokens,'' )
    cString := Substr(cString,2)
  ELSEIF nFound > 0
    AADD( aTokens, SubStr(cString,1,nFound-1) )
    cString := SubStr(cString,nFound+1)
  ELSE
    AADD( aTokens, cString )
    EXIT
  ENDIF
ENDDO
RETURN aTokens

* ----------

STATIC FUNCTION _tokendelims

RETURN ' ,.;:!?/()#&%+-*'+CHR(0)+CHR(9)+CHR(10)+CHR(13)+CHR(26)


